import firebase from 'firebase/app';
import 'firebase/firestore';
import { DocumentData } from './utils';
import 'core-js/features/array/includes';
import 'core-js/features/number/is-nan';
export declare const DOCUMENT_KEY_NAME = "__name__";
export declare class FireSQL {
    private _ref?;
    private _path?;
    private _options;
    constructor(refOrOptions?: FirestoreOrDocOrOptions, options?: FireSQLOptions);
    readonly ref: firebase.firestore.DocumentReference;
    readonly firestore: firebase.firestore.Firestore;
    readonly options: FireSQLOptions;
    query(sql: string, options?: QueryOptions): Promise<DocumentData[]>;
    query<T>(sql: string, options?: QueryOptions): Promise<T[]>;
    private _getRef;
    toJSON(): object;
}
export interface FireSQLOptions {
    includeId?: boolean | string;
}
export interface QueryOptions extends FireSQLOptions {
}
declare type FirestoreOrDocument = string | firebase.firestore.Firestore | firebase.firestore.DocumentReference | AdminFirestore | AdminDocumentReference;
export declare type FirestoreOrDocOrOptions = FirestoreOrDocument | FireSQLOptions;
/**
 * An interface representing the basics we need from the
 * admin.firestore.Firestore class.
 * We use it like this to avoid having to require "firebase-admin".
 */
interface AdminFirestore {
    collection(collectionPath: string): any;
    doc(documentPath: string): any;
}
/**
 * An interface representing the basics we need from the
 * admin.firestore.DocumentReference class.
 * We use it like this to avoid having to require "firebase-admin".
 */
interface AdminDocumentReference {
    collection(collectionPath: string): any;
    get(options?: any): Promise<any>;
}
/**
 * @deprecated Class FirestoreSQL has been renamed FireSQL
 */
export declare class FirestoreSQL extends FireSQL {
    constructor(ref?: string | firebase.firestore.Firestore | firebase.firestore.DocumentReference);
}
export {};
